@@ -17,25 +17,33 @@
 #
 # author Salvo "LtWorf" Tomaselli <tiposchi@tiscali.it>
 
+import atexit
 import datetime
 from enum import Enum
+from pathlib import Path
 import re
 import select
+import signal
 import socket
 import argparse
 from typing import *
+import os
 from os import environ
 from os.path import expanduser
+import pwd
 from socket import gethostname
+import sys
 import traceback
 
 import slack
 import rocket
+from log import *
 
 
 # How slack expresses mentioning users
 _MENTIONS_REGEXP = re.compile(r'<@([0-9A-Za-z]+)>')
 _CHANNEL_MENTIONS_REGEXP = re.compile(r'<#[A-Z0-9]+\|([A-Z0-9\-a-z]+)>')
+_URL_REGEXP=re.compile(r'<([a-z0-9\-\.]+)://([^\s\|]+)[\|]{0,1}([^<>]*)>')
 
 
 _SLACK_SUBSTITUTIONS = [
@@ -47,6 +55,7 @@ _SLACK_SUBSTITUTIONS = [
 
 class Replies(Enum):
     RPL_LUSERCLIENT = 251
+    RPL_USERHOST = 302
     RPL_UNAWAY = 305
     RPL_NOWAWAY = 306
     RPL_ENDOFWHO = 315
@@ -73,7 +82,7 @@ MPIM_HIDE_DELAY = datetime.timedelta(days=50)
 
 
 class Client:
-    def __init__(self, s, sl_client, nouserlist: bool, autojoin: bool, provider: Provider):
+    def __init__(self, s, sl_client: Union[slack.Slack, rocket.Rocket], nouserlist: bool, autojoin: bool, provider: Provider):
         self.nick = b''
         self.username = b''
         self.realname = b''
@@ -85,6 +94,8 @@ class Client:
         self.sl_client = sl_client
         self.nouserlist = nouserlist
         self.autojoin = autojoin
+        self._usersent = False # Used to hold all events until the IRC client sends the initial USER message
+        self._held_events: List[slack.SlackEvent] = []
 
         if self.provider == Provider.SLACK:
             self.substitutions = _SLACK_SUBSTITUTIONS
@@ -94,10 +105,11 @@ class Client:
     def _nickhandler(self, cmd: bytes) -> None:
         _, nick = cmd.split(b' ', 1)
         self.nick = nick.strip()
+        assert self.sl_client.login_info
         if self.nick != self.sl_client.login_info.self.name.encode('ascii'):
             self._sendreply(Replies.ERR_ERRONEUSNICKNAME, 'Incorrect nickname, use %s' % self.sl_client.login_info.self.name)
 
-    def _sendreply(self, code: Union[int,Replies], message: Union[str,bytes], extratokens: List[Union[str,bytes]] = []) -> None:
+    def _sendreply(self, code: Union[int,Replies], message: Union[str,bytes], extratokens: Iterable[Union[str,bytes]] = []) -> None:
         codeint = code if isinstance(code, int) else code.value
         bytemsg = message if isinstance(message, bytes) else message.encode('utf8')
 
@@ -115,6 +127,7 @@ class Client:
 
     def _userhandler(self, cmd: bytes) -> None:
         #TODO USER salvo 8 * :Salvatore Tomaselli
+        assert self.sl_client.login_info
         self._sendreply(1, 'Welcome to localslackirc')
         self._sendreply(2, 'Your team name is: %s' % self.sl_client.login_info.team.name)
         self._sendreply(2, 'Your team domain is: %s' % self.sl_client.login_info.team.domain)
@@ -144,31 +157,47 @@ class Client:
                 channel_name = '#%s' % sl_chan.name_normalized
                 self.parted_channels.add(channel_name.encode('utf-8'))
 
+        # Eventual channel joining done, sending the held events
+        self._usersent = True
+        for ev in self._held_events:
+            self.slack_event(ev)
+        self._held_events = []
 
     def _pinghandler(self, cmd: bytes) -> None:
         _, lbl = cmd.split(b' ', 1)
         self.s.send(b':%s PONG %s %s\n' % (self.hostname, self.hostname, lbl))
 
     def _joinhandler(self, cmd: bytes) -> None:
-        _, channel_name = cmd.split(b' ', 1)
+        _, channel_name_b = cmd.split(b' ', 1)
 
-        if channel_name in self.parted_channels:
-            self.parted_channels.remove(channel_name)
+        if channel_name_b in self.parted_channels:
+            self.parted_channels.remove(channel_name_b)
 
+        channel_name = channel_name_b[1:].decode()
         try:
-            slchan = self.sl_client.get_channel_by_name(channel_name[1:].decode())
-        except:
+            slchan = self.sl_client.get_channel_by_name(channel_name)
+        except Exception:
+            self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to find channel: {channel_name}')
             return
 
-        self._send_chan_info(channel_name, slchan)
+        if not slchan.is_member:
+            try:
+                self.sl_client.join(slchan)
+            except Exception:
+                self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to join server channel: {channel_name}')
+
+        try:
+            self._send_chan_info(channel_name_b, slchan)
+        except Exception:
+            self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to join channel: {channel_name}')
 
     def _send_chan_info(self, channel_name: bytes, slchan: slack.Channel):
         if not self.nouserlist:
-            userlist = []  # type List[bytes]
+            userlist: List[bytes] = []
             for i in self.sl_client.get_members(slchan.id):
                 try:
                     u = self.sl_client.get_user(i)
-                except:
+                except Exception:
                     continue
                 if u.deleted:
                     # Disabled user, skip it
@@ -213,11 +242,11 @@ class Client:
                     message,
                     action,
                 )
-            except:
-                print('Impossible to find user ', dest)
+            except Exception:
+                log('Impossible to find user ', dest)
 
     def _listhandler(self, cmd: bytes) -> None:
-        for c in self.sl_client.channels():
+        for c in self.sl_client.channels(refresh=True):
             self._sendreply(Replies.RPL_LIST, c.real_topic, ['#' + c.name, str(c.num_members)])
         self._sendreply(Replies.RPL_LISTEND, 'End of LIST')
 
@@ -261,22 +290,35 @@ class Client:
         self._sendreply(response, 'Away status changed')
 
     def _topichandler(self, cmd: bytes) -> None:
-        _, channel, topic = cmd.split(b' ', 2)
-        channel = self.sl_client.get_channel_by_name(channel.decode()[1:])
-        self.sl_client.topic(channel, topic.decode()[1:])
+        _, channel_b, topic_b = cmd.split(b' ', 2)
+        topic = topic_b.decode()[1:]
+        channel = self.sl_client.get_channel_by_name(channel_b.decode()[1:])
+        try:
+            self.sl_client.topic(channel, topic)
+        except Exception:
+            self._sendreply(Replies.ERR_UNKNOWNCOMMAND, f'Unable to set topic to {topic}')
 
     def _kickhandler(self, cmd: bytes) -> None:
-        _, channel, username, message = cmd.split(b' ', 3)
-        channel = self.sl_client.get_channel_by_name(channel.decode()[1:])
+        _, channel_b, username, message = cmd.split(b' ', 3)
+        channel = self.sl_client.get_channel_by_name(channel_b.decode()[1:])
         user = self.sl_client.get_user_by_name(username.decode())
         try:
             self.sl_client.kick(channel, user)
         except Exception as e:
             self._sendreply(Replies.ERR_UNKNOWNCOMMAND, 'Error: %s' % e)
 
+    def _userhosthandler(self, cmd: bytes) -> None:
+        nicknames = cmd.split(b' ')
+        del nicknames[0] # Remove the command itself
+        #TODO replace + with - in case of away
+        #TODO append a * to the nickname for OP
+
+        replies = (b'%s=+unknown' % i for i in nicknames)
+        self._sendreply(Replies.RPL_USERHOST, '', replies)
+
     def _invitehandler(self, cmd: bytes) -> None:
-        _, username, channel = cmd.split(b' ', 2)
-        channel = self.sl_client.get_channel_by_name(channel.decode()[1:])
+        _, username, channel_b = cmd.split(b' ', 2)
+        channel = self.sl_client.get_channel_by_name(channel_b.decode()[1:])
         user = self.sl_client.get_user_by_name(username.decode())
         try:
             self.sl_client.invite(channel, user)
@@ -286,8 +328,13 @@ class Client:
     def _whohandler(self, cmd: bytes) -> None:
         _, name = cmd.split(b' ', 1)
         if not name.startswith(b'#'):
-            print('WHO not supported on ', name)
+            try:
+                user = self.sl_client.get_user_by_name(name.decode())
+            except KeyError:
+                return
+            self._sendreply(Replies.RPL_WHOREPLY, '0 %s' % user.real_name, [name, user.name, '127.0.0.1', self.hostname, user.name, 'H'])
             return
+
         try:
             channel = self.sl_client.get_channel_by_name(name.decode()[1:])
         except KeyError:
@@ -297,7 +344,7 @@ class Client:
             try:
                 user = self.sl_client.get_user(i)
                 self._sendreply(Replies.RPL_WHOREPLY, '0 %s' % user.real_name, [name, user.name, '127.0.0.1', self.hostname, user.name, 'H'])
-            except:
+            except Exception:
                 pass
         self._sendreply(Replies.RPL_ENDOFWHO, 'End of WHO list', [name])
 
@@ -318,9 +365,7 @@ class Client:
         if self.provider == Provider.SLACK:
             msg = msg.replace('@here', '<!here>')
             msg = msg.replace('@channel', '<!channel>')
-            msg = msg.replace('@yell', '<!channel>')
-            msg = msg.replace('@shout', '<!channel>')
-            msg = msg.replace('@attention', '<!channel>')
+            msg = msg.replace('@everyone', '<!everyone>')
         elif self.provider == Provider.ROCKETCHAT:
             msg = msg.replace('@yell', '@channel')
             msg = msg.replace('@shout', '@channel')
@@ -367,6 +412,18 @@ class Client:
                         i[mention.span()[1]:]
                     )
 
+                while True:
+                    url = _URL_REGEXP.search(i)
+                    if not url:
+                        break
+                    schema, path, label = url.groups()
+                    i = (
+                        i[0:url.span()[0]] +
+                        f'{schema}://{path}' +
+                        (f' ({label})' if label else '') +
+                        i[url.span()[1]:]
+                    )
+
             for s in self.substitutions:
                 i = i.replace(s[0], s[1])
 
@@ -375,6 +432,7 @@ class Client:
             if self.provider == Provider.SLACK:
                 encoded = encoded.replace(b'<!here>', b'yelling [%s]' % self.nick)
                 encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER [%s]' % self.nick)
+                encoded = encoded.replace(b'<!everyone>', b'DEAFENING YELL [%s]' % self.nick)
             elif self.provider == Provider.ROCKETCHAT:
                 encoded = encoded.replace(b'@here', b'yelling [%s]' % self.nick)
                 encoded = encoded.replace(b'@channel', b'YELLING LOUDER [%s]' % self.nick)
@@ -395,7 +453,7 @@ class Client:
         except KeyError:
             dest = self.nick
         except Exception as e:
-            print('Error: ', str(e))
+            log('Error: ', str(e))
             return
         if dest in self.parted_channels:
             # Ignoring messages, channel was left on IRC
@@ -429,6 +487,9 @@ class Client:
             self.s.send(b':%s!%s@127.0.0.1 PART %s\n' % (name, rname, dest))
 
     def slack_event(self, sl_ev: slack.SlackEvent) -> None:
+        if not self._usersent:
+            self._held_events.append(sl_ev)
+            return
         if isinstance(sl_ev, slack.MessageDelete):
             self._message(sl_ev, '[deleted]')
         elif isinstance(sl_ev, slack.Message):
@@ -449,6 +510,9 @@ class Client:
             self._joined_parted(sl_ev, False)
         elif isinstance(sl_ev, slack.TopicChange):
             self._sendreply(Replies.RPL_TOPIC, sl_ev.topic, ['#' + self.sl_client.get_channel(sl_ev.channel).name])
+        elif isinstance(sl_ev, slack.GroupJoined):
+            channel_name = '#%s' % sl_ev.channel.name_normalized
+            self._send_chan_info(channel_name.encode('utf-8'), sl_ev.channel)
 
     def command(self, cmd: bytes) -> None:
         if b' ' in cmd:
@@ -473,7 +537,7 @@ class Client:
             b'sendfile': self._sendfilehandler,
             #QUIT
             #CAP LS
-            #USERHOST
+            b'USERHOST': self._userhosthandler,
             #Unknown command:  b'whois TAMARRO'
         }
 
@@ -481,14 +545,43 @@ class Client:
             handlers[cmdid](cmd)
         else:
             self._sendreply(Replies.ERR_UNKNOWNCOMMAND, 'Unknown command', [cmdid])
-            print('Unknown command: ', cmd)
+            log('Unknown command: ', cmd)
+
+
+def exit_hook(status_file, sl_client) -> None:
+    if status_file:
+        log(f'Writing status to {status_file}')
+        status_file.write_bytes(sl_client.get_status())
+    log('Exiting...')
+
+
+def su() -> None:
+    """
+    switch user. Useful when starting localslackirc
+    as a service as root user.
+    """
+    if sys.platform.startswith('win'):
+        return
+
+    # Nothing to do, already not root
+    if os.getuid() != 0:
+        return
+
+    username = environ.get('PROCESS_OWNER', 'nobody')
+    userdata = pwd.getpwnam(username)
+    os.setgid(userdata.pw_gid)
+    os.setegid(userdata.pw_gid)
+    os.setuid(userdata.pw_uid)
+    os.seteuid(userdata.pw_uid)
 
 
 def main() -> None:
+    su()
+
     parser = argparse.ArgumentParser()
     parser.add_argument('-p', '--port', type=int, action='store', dest='port',
                                 default=9007, required=False,
-                                help='set port number')
+                                help='set port number. Defaults to 9007')
     parser.add_argument('-i', '--ip', type=str, action='store', dest='ip',
                                 default='127.0.0.1', required=False,
                                 help='set ip address')
@@ -496,6 +589,10 @@ def main() -> None:
                                 default=expanduser('~')+'/.localslackirc',
                                 required=False,
                                 help='set the token file')
+    parser.add_argument('-c', '--cookiefile', type=str, action='store', dest='cookiefile',
+                                default=None,
+                                required=False,
+                                help='set the cookie file (for slack only, for xoxc tokens)')
     parser.add_argument('-u', '--nouserlist', action='store_true',
                                 dest='nouserlist', required=False,
                                 help='don\'t display userlist')
@@ -507,18 +604,35 @@ def main() -> None:
                                 help='allow non 127. addresses, this is potentially dangerous')
     parser.add_argument('--rc-url', type=str, action='store', dest='rc_url', default=None, required=False,
                                 help='The rocketchat URL. Setting this changes the mode from slack to rocketchat')
+    parser.add_argument('-f', '--status-file', type=str, action='store', dest='status_file', required=False, default=None,
+                                help='Path to the file to keep the internal status.')
+    parser.add_argument('--log-suffix', type=str, action='store', dest='log_suffix', default='',
+                                help='Set a suffix for the syslog identifier')
 
     args = parser.parse_args()
+
+    openlog(environ.get('LOG_SUFFIX', args.log_suffix))
+
+    status_file_str: Optional[str] = environ.get('STATUS_FILE', args.status_file)
+    status_file = None
+    if status_file_str is not None:
+        log('Status file at:', status_file_str)
+        status_file = Path(status_file_str)
+
+    ip: str = environ.get('IP_ADDRESS', args.ip)
+    overridelocalip: bool = environ['OVERRIDE_LOCAL_IP'].lower() == 'true' if 'OVERRIDE_LOCAL_IP' in environ else args.overridelocalip
+
     # Exit if their chosden ip isn't local. User can override with -o if they so dare
-    if not args.ip.startswith('127') and not args.overridelocalip:
+    if not ip.startswith('127') and not overridelocalip:
         exit('supplied ip isn\'t local\nlocalslackirc has no encryption or ' \
                 'authentication, it\'s recommended to only allow local connections\n' \
                 'you can override this with -o')
 
-    if 'PORT' in environ:
-        port = int(environ['PORT'])
-    else:
-        port = args.port
+    port = int(environ.get('PORT', args.port))
+    rc_url: Optional[str] = environ.get('RC_URL', args.rc_url)
+
+    autojoin: bool = environ['AUTOJOIN'].lower() == 'true' if 'AUTOJOIN' in environ else args.autojoin
+    nouserlist: bool = environ['NOUSERLIST'].lower() == 'true' if 'NOUSERLIST' in environ else args.nouserlist
 
     if 'TOKEN' in environ:
         token = environ['TOKEN']
@@ -526,26 +640,55 @@ def main() -> None:
         try:
             with open(args.tokenfile) as f:
                 token = f.readline().strip()
+        except IsADirectoryError:
+            exit(f'Not a file {args.tokenfile}')
         except (FileNotFoundError, PermissionError):
             exit(f'Unable to open the token file {args.tokenfile}')
 
-    if args.rc_url:
-        sl_client: Union[slack.Slack, rocket.Rocket] = rocket.Rocket(args.rc_url, token)
+    if 'COOKIE' in environ:
+        cookie: Optional[str] = environ['COOKIE']
+    else:
+        try:
+            if args.cookiefile:
+                with open(args.cookiefile) as f:
+                    cookie = f.readline().strip()
+            else:
+                cookie = None
+        except (FileNotFoundError, PermissionError):
+            exit(f'Unable to open the cookie file {args.cookiefile}')
+        except IsADirectoryError:
+            exit(f'Not a file {args.cookiefile}')
+
+    if token.startswith('xoxc-') and not cookie:
+        exit('The cookie is needed for this kind of slack token')
+
+    previous_status = None
+    if status_file is not None and status_file.exists():
+        previous_status = status_file.read_bytes()
+
+    if rc_url:
+        sl_client: Union[slack.Slack, rocket.Rocket] = rocket.Rocket(rc_url, token, previous_status)
         provider = Provider.ROCKETCHAT
     else:
-        sl_client = slack.Slack(token)
+        sl_client = slack.Slack(token, cookie, previous_status)
         provider = Provider.SLACK
+
+    atexit.register(exit_hook, status_file, sl_client)
+    term_f = lambda _, __: sys.exit(0)
+    signal.signal(signal.SIGHUP, term_f )
+    signal.signal(signal.SIGTERM, term_f)
+
     sl_events = sl_client.events_iter()
     serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
     serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-    serversocket.bind((args.ip, port))
+    serversocket.bind((ip, port))
     serversocket.listen(1)
 
     poller = select.poll()
 
     while True:
         s, _ = serversocket.accept()
-        ircclient = Client(s, sl_client, args.nouserlist, args.autojoin, provider)
+        ircclient = Client(s, sl_client, nouserlist, autojoin, provider)
 
         poller.register(s.fileno(), select.POLLIN)
         if sl_client.fileno is not None:
@@ -568,7 +711,7 @@ def main() -> None:
                         ircclient.command(i)
 
             while sl_event:
-                print(sl_event)
+                log(sl_event)
                 ircclient.slack_event(sl_event)
                 sl_event = next(sl_events)
 
