@@ -17,36 +17,35 @@
 # author Salvo "LtWorf" Tomaselli <tiposchi@tiscali.it>
 
 import datetime
+from dataclasses import dataclass, field
 from functools import lru_cache
+import json
 from time import sleep, time
 from typing import *
 
-from slackclient import SlackClient
 from typedload import load, dump
 
 from diff import seddiff
+from slackclient import SlackClient
+from slackclient.client import LoginInfo
+from log import *
 
 
 USELESS_EVENTS = {
     'channel_marked',
     'group_marked',
+    'mpim_marked',
     'hello',
     'dnd_updated_user',
     'reaction_added',
     'user_typing',
     'file_deleted',
     'file_public',
+    'file_created',
+    'desktop_notification',
 }
 
 
-def _loadwrapper(value, type_):
-    try:
-        return load(value, type_)
-    except Exception as e:
-        print(e)
-        pass
-
-
 class ResponseException(Exception):
     pass
 
@@ -115,13 +114,28 @@ class Message(NamedTuple):
     text: str
 
 
+class NoChanMessage(NamedTuple):
+    user: str
+    text: str
+
+
 class ActionMessage(Message):
     pass
 
 
-class MessageEdit(NamedTuple):
-    previous: Message
-    current: Message
+@dataclass
+class GroupJoined:
+    type: Literal['group_joined']
+    channel: Channel
+
+
+@dataclass
+class MessageEdit:
+    type: Literal['message']
+    subtype: Literal['message_changed']
+    channel: str
+    previous: NoChanMessage = field(metadata={'name': 'previous_message'})
+    current: NoChanMessage = field(metadata={'name': 'message'})
 
     @property
     def is_changed(self) -> bool:
@@ -129,13 +143,27 @@ class MessageEdit(NamedTuple):
 
     @property
     def diffmsg(self) -> Message:
-        m = dump(self.current)
-        m['text'] = seddiff(self.previous.text, self.current.text)
-        return load(m, Message)
+        return Message(
+            text=seddiff(self.previous.text, self.current.text),
+            channel=self.channel,
+            user=self.current.user,
+        )
 
 
-class MessageDelete(Message):
-    pass
+@dataclass
+class MessageDelete:
+    type: Literal['message']
+    subtype: Literal['message_deleted']
+    channel: str
+    previous_message: NoChanMessage
+
+    @property
+    def user(self) -> str:
+        return self.previous_message.user
+
+    @property
+    def text(self) -> str:
+        return self.previous_message.text
 
 
 class Profile(NamedTuple):
@@ -146,7 +174,8 @@ class Profile(NamedTuple):
     is_ultra_restricted: bool = False
 
 
-class File(NamedTuple):
+@dataclass
+class File:
     id: str
     url_private: str
     size: int
@@ -154,9 +183,9 @@ class File(NamedTuple):
     name: Optional[str] = None
     title: Optional[str] = None
     mimetype: Optional[str] = None
-    channels: List[str] = list()
-    groups: List[str] = list()
-    ims: List[str] = list()
+    channels: List[str] = field(default_factory=list)
+    groups: List[str] = field(default_factory=list)
+    ims: List[str] = field(default_factory=list)
 
     def announce(self) -> Message:
         """
@@ -174,13 +203,18 @@ class File(NamedTuple):
         )
 
 
-class FileShared(NamedTuple):
+@dataclass
+class FileShared:
+    type: Literal['file_shared']
     file_id: str
     user_id: str
     ts: float
 
 
-class MessageBot(NamedTuple):
+@dataclass
+class MessageBot:
+    type: Literal['message']
+    subtype: Literal['bot_message']
     text: str
     username: str
     channel: str
@@ -205,20 +239,59 @@ class IM(NamedTuple):
 
 
 class Join(NamedTuple):
+    type: Literal['member_joined_channel']
     user: str
     channel: str
 
 
 class Leave(NamedTuple):
+    type: Literal['member_left_channel']
     user: str
     channel: str
 
 
-class TopicChange(NamedTuple):
+@dataclass
+class TopicChange:
+    type: Literal['message']
+    subtype: Literal['group_topic']
     topic: str
     channel: str
     user: str
 
+
+@dataclass
+class HistoryBotMessage:
+    type: Literal['message']
+    subtype: Literal['bot_message']
+    text: str
+    bot_id: Optional[str]
+    username: str = 'bot'
+    ts: float = 0
+    files: List[File] = field(default_factory=list)
+    thread_ts: Optional[str] = None
+
+
+@dataclass
+class HistoryMessage:
+    type: Literal['message']
+    user: str
+    text: str
+    ts: float
+    files: List[File] = field(default_factory=list)
+    thread_ts: Optional[str] = None
+
+
+class NextCursor(NamedTuple):
+    next_cursor: str
+
+
+class History(NamedTuple):
+    ok: Literal[True]
+    messages: List[Union[HistoryMessage, HistoryBotMessage]]
+    has_more: bool
+    response_metadata: Optional[NextCursor] = None
+
+
 SlackEvent = Union[
     TopicChange,
     MessageDelete,
@@ -229,12 +302,22 @@ SlackEvent = Union[
     FileShared,
     Join,
     Leave,
+    GroupJoined,
 ]
 
 
+@dataclass
+class SlackStatus:
+    """
+    Not related to the slack API.
+    This is a structure used internally by this module to
+    save the status on disk.
+    """
+    last_timestamp: float = 0.0
+
 class Slack:
-    def __init__(self, token: str) -> None:
-        self.client = SlackClient(token)
+    def __init__(self, token: str, cookie: Optional[str], previous_status: Optional[bytes]) -> None:
+        self.client = SlackClient(token, cookie)
         self._usercache: Dict[str, User] = {}
         self._usermapcache: Dict[str, User] = {}
         self._imcache: Dict[str, str] = {}
@@ -242,6 +325,126 @@ class Slack:
         self._get_members_cache_cursor: Dict[str, Optional[str]] = {}
         self._internalevents: List[SlackEvent] = []
         self._sent_by_self: Set[float] = set()
+        self.login_info: Optional[LoginInfo] = None
+        if previous_status is None:
+            self._status = SlackStatus()
+        else:
+            self._status = load(json.loads(previous_status), SlackStatus)
+
+    def _thread_history(self, channel: str, thread_id: str) -> List[Union[HistoryMessage, HistoryBotMessage]]:
+        r: List[Union[HistoryMessage, HistoryBotMessage]] = []
+        cursor = None
+        log('Thread history', channel, thread_id)
+        while True:
+            log('Cursor')
+            p = self.client.api_call(
+                'conversations.replies',
+                channel=channel,
+                ts=thread_id,
+                limit=1000,
+                cursor=cursor,
+            )
+            try:
+                response = load(p, History)
+            except Exception as e:
+                    log('Failed to parse', e)
+                    log(p)
+                    break
+            r += [i for i in response.messages if i.ts != i.thread_ts]
+            if response.has_more and response.response_metadata:
+                cursor = response.response_metadata.next_cursor
+            else:
+                break
+        log('Thread fetched')
+        r[0].thread_ts = None
+        return r
+
+    def _history(self) -> None:
+        '''
+        Obtain the history from the last known event and
+        inject fake events as if the messages are coming now.
+        '''
+        log('Fetching history...')
+
+        if self._status.last_timestamp == 0:
+            log('No last known timestamp. Unable to fetch history')
+            return
+
+        last_timestamp = self._status.last_timestamp
+        dt = datetime.datetime.fromtimestamp(last_timestamp)
+        log(f'Last known timestamp {dt}')
+
+        for channel in self.channels():
+            if not channel.is_member:
+                continue
+            log(f'Downloading logs from channel {channel.name_normalized}')
+
+            cursor = None
+            while True: # Loop to iterate the cursor
+                log('Calling cursor')
+                r = self.client.api_call(
+                    'conversations.history',
+                    channel=channel.id,
+                    oldest=last_timestamp,
+                    limit=1000,
+                    cursor=cursor,
+                )
+                try:
+                    response = load(r, History)
+                except Exception as e:
+                    log('Failed to parse', e)
+                    log(r)
+                    break
+                msg_list = list(response.messages)
+                while msg_list:
+                    msg = msg_list.pop(0)
+
+                    # The last seen message is sent again, skip it
+                    if msg.ts == last_timestamp:
+                        continue
+                    # Update the last seen timestamp
+                    if self._status.last_timestamp < msg.ts:
+                        self._status.last_timestamp = msg.ts
+
+                    # The attached files
+                    for f in msg.files:
+                        f.channels.append(channel.id)
+                        self._internalevents.append(f.announce())
+
+                    # History for the thread
+                    if  msg.thread_ts and float(msg.thread_ts) == msg.ts:
+                        l = self._thread_history(channel.id, msg.thread_ts)
+                        l.reverse()
+                        msg_list = l + msg_list
+                        continue
+
+                    # Inject the events
+                    if isinstance(msg, HistoryMessage):
+                        self._internalevents.append(Message(
+                            channel=channel.id,
+                            text=msg.text,
+                            user=msg.user
+                        ))
+                    elif isinstance(msg, HistoryBotMessage):
+                        self._internalevents.append(MessageBot(
+                            type='message', subtype='bot_message',
+                            text=msg.text,
+                            username=msg.username,
+                            channel=channel.id,
+                            bot_id=msg.bot_id,
+                        ))
+
+                if response.has_more and response.response_metadata:
+                    cursor = response.response_metadata.next_cursor
+                else:
+                    break
+
+    def get_status(self) -> bytes:
+        '''
+        A status string that will be passed back when this is started again
+        '''
+        return json.dumps(dump(self._status), ensure_ascii=True).encode('ascii')
+
 
     def away(self, is_away: bool) -> None:
         """
@@ -265,6 +468,12 @@ class Slack:
         if not response.ok:
             raise ResponseException(response)
 
+    def join(self, channel: Channel) -> None:
+        r = self.client.api_call('conversations.join', channel=channel.id)
+        response = load(r, Response)
+        if not response.ok:
+            raise ResponseException(response)
+
     def invite(self, channel: Channel, user: Union[User, List[User]]) -> None:
         if isinstance(user, User):
             ids = user.id
@@ -306,17 +515,14 @@ class Slack:
         # Generate all the Join events, if this is not the 1st iteration
         if id_ in self._get_members_cache:
             for i in newusers.difference(cached):
-                self._internalevents.append(Join(user=i, channel=id_))
+                self._internalevents.append(Join('member_joined_channel', user=i, channel=id_))
 
         self._get_members_cache[id_] = cached.union(newusers)
         self._get_members_cache_cursor[id_] = r.get('response_metadata', {}).get('next_cursor')
         return self._get_members_cache[id_]
 
     @lru_cache()
-    def channels(self) -> List[Channel]:
-        """
-        Returns the list of slack channels
-        """
+    def _channels(self) -> List[Channel]:
         result: List[Channel] = []
         r = self.client.api_call("conversations.list", exclude_archived=True,
                 types='public_channel,private_channel,mpim', limit=1000)
@@ -326,6 +532,15 @@ class Slack:
         else:
             raise ResponseException(response)
 
+    def channels(self, refresh: bool = False) -> List[Channel]:
+        """
+        Returns the list of slack channels
+
+        if refresh is set, the local cache is cleared
+        """
+        if refresh:
+            self._channels.cache_clear()
+        return self._channels()
 
     @lru_cache()
     def get_channel(self, id_: str) -> Channel:
@@ -334,9 +549,10 @@ class Slack:
 
         raises KeyError if it doesn't exist.
         """
-        for c in self.channels():
-            if c.id == id_:
-                return c
+        for i in range(2):
+            for c in self.channels(refresh=bool(i)):
+                if c.id == id_:
+                    return c
         raise KeyError()
 
     @lru_cache()
@@ -346,9 +562,10 @@ class Slack:
 
         raises KeyError if it doesn't exist.
         """
-        for c in self.channels():
-            if c.name == name:
-                return c
+        for i in range(2):
+            for c in self.channels(refresh=bool(i)):
+                if c.name == name:
+                    return c
         raise KeyError()
 
     @property
@@ -532,32 +749,58 @@ class Slack:
 
             try:
                 events = self.client.rtm_read()
-            except:
-                print('Connecting to slack...')
+            except Exception:
+                events = []
+                log('Connecting to slack...')
                 try:
                     self.login_info = self.client.rtm_connect()
                     sleeptime = 1
+                    self._history()
                 except Exception as e:
-                    print(f'Connection to slack failed {e}')
+                    log(f'Connection to slack failed {e}')
                     sleep(sleeptime)
                     if sleeptime <= 120:  # max reconnection interval at 2 minutes
                         sleeptime *= 2
                     continue
-                print('Connected to slack')
+                log('Connected to slack')
                 continue
 
             for event in events:
                 t = event.get('type')
-                subt = event.get('subtype')
-
                 ts = float(event.get('ts', 0))
+
+                if ts > self._status.last_timestamp:
+                    self._status.last_timestamp = ts
+
                 if ts in self._sent_by_self:
                     self._sent_by_self.remove(ts)
                     continue
 
+                if t in USELESS_EVENTS:
+                        continue
+
+                try:
+                    ev = load(
+                        event,
+                        Union[TopicChange, FileShared, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave]
+                    )
+                except Exception:
+                    ev = None
+
+                if isinstance(ev, Join):
+                    self._get_members_cache[ev.channel].add(ev.user)
+                elif isinstance(ev, Leave):
+                    self._get_members_cache[ev.channel].remove(ev.user)
+
+                if ev:
+                    yield ev
+
+
+                subt = event.get('subtype')
+
                 try:
                     if t == 'message' and (not subt or subt == 'me_message'):
-                        msg = _loadwrapper(event, Message)
+                        msg = load(event, Message)
 
                         # In private chats, pretend that my own messages
                         # sent from another client actually come from
@@ -570,31 +813,7 @@ class Slack:
                         else:
                             yield msg
                     elif t == 'message' and subt == 'slackbot_response':
-                        yield _loadwrapper(event, Message)
-                    elif t == 'message' and subt == 'message_changed':
-                        event['message']['channel'] = event['channel']
-                        event['previous_message']['channel'] = event['channel']
-                        yield MessageEdit(
-                            previous=load(event['previous_message'], Message),
-                            current=load(event['message'], Message)
-                        )
-                    elif t == 'message' and subt == 'group_topic':
-                        yield _loadwrapper(event, TopicChange)
-                    elif t == 'message' and subt == 'message_deleted':
-                        event['previous_message']['channel'] = event['channel']
-                        ev = _loadwrapper(event['previous_message'], MessageDelete)
-                        if ev.text:  # deleting files generates empty MessageDelete events
-                            yield ev
-                    elif t == 'message' and subt == 'bot_message':
-                        yield _loadwrapper(event, MessageBot)
-                    elif t == 'member_joined_channel':
-                        j = _loadwrapper(event, Join)
-                        self._get_members_cache[j.channel].add(j.user)
-                        yield j
-                    elif t == 'member_left_channel':
-                        j = _loadwrapper(event, Leave)
-                        self._get_members_cache[j.channel].remove(j.user)
-                        yield j
+                        yield load(event, Message)
                     elif t == 'user_change':
                         # Changes in the user, drop it from cache
                         u = load(event['user'], User)
@@ -602,16 +821,9 @@ class Slack:
                             del self._usercache[u.id]
                             #FIXME don't know if it is wise, maybe it gets lost forever del self._usermapcache[u.name]
                         #TODO make an event for this
-                    elif t == 'file_shared':
-                        try: # slack idiocy workaround, they send the event twice, one time without the ts field
-                            yield _loadwrapper(event, FileShared)
-                        except ValueError:
-                            pass
-                    elif t in USELESS_EVENTS:
-                        continue
                     else:
-                        print(event)
+                        log(event)
                 except Exception as e:
-                    print('Exception: %s' % e)
+                    log('Exception: %s' % e)
             self._triage_sent_by_self()
             yield None
